#ifndef _RANGELIST_H_
#define _RANGELIST_H_

#include <vector>
#include <set>
#include <map>
#include <algorithm>
#include <cassert>

#include "TypeConversion.h"
#include "Utils.h"
#include "StringArray.h"
#include "InputFile.h"

/**
   range is inclusive on both edges.
 */
struct PositionPair{
	 std::string chr;
    unsigned int begin;
    unsigned int end;
PositionPair():
     chr(""), begin(0), end(0) 
        {};
PositionPair(std::string c, unsigned int b, unsigned int e):
    chr(c), begin(b), end(e) 
        {};
    bool operator== (const PositionPair o){
        return (begin == o.begin && end == o.end);
    }
    bool isInRange(const unsigned int i){
        return (this->begin <= i && i <= this->end);
    }
};
bool PositionPairCompare(const PositionPair& p1, const PositionPair& p2){
    if (p1.begin != p2.begin)
        return (p1.begin < p2.begin);
    return (p1.end < p2.end);
}


class RangeCollection{
  public:
    void addRange(const std::string& uniq_id, const std::string& chr, unsigned int begin, unsigned int end) {
        // if chr not exists
        // add chr to the chrVector
        std::string c = chopChr(chr);
        if (this->rangeMap.find(uniq_id) == this->rangeMap.end()){
            chrVector.push_back(uniq_id);
        }
        
        // add begin, end to that chr entry
        this->rangeMap[uniq_id].push_back( PositionPair(chr, begin, end) );
    }
    void sort() {
        /*
         * some test code
         */
        /*
        this->addRange("X", 2, 4 );
        this->addRange("X", 1, 3 );
        this->addRange("1", 1, 3 );
        this->addRange("1", 4, 6 );
        this->addRange("1", 5, 10 );
        */
        this->sortChrVector();
        this->consolidate();
        /*
        assert(chrVector[0] == "1");
        assert(rangeMap["1"][0] == PositionPair(1, 3));
        assert(rangeMap["1"][1] == PositionPair(4, 10));
        assert(rangeMap["X"][0] == PositionPair(1, 4));
        */
    }
    
    void obtainRange(const unsigned int index, std::string* range) {
        assert(range);
        range->clear();
        int t = index; 
        int s; 
        for(unsigned int i = 0; i < this->chrVector.size(); i++ ) {
            s = rangeMap[chrVector[i]].size();
            if ( t < s) {
                (*range) = chrVector[i];
                range->push_back(':');
                (*range) += toString((this->rangeMap[this->chrVector[i]][t]).begin);
                range->push_back('-');
                (*range) += toString((this->rangeMap[this->chrVector[i]][t]).end);
                return;
            } else {
                t -= s;
            }
        }
        assert(false);
    };   
    void obtainRange(const unsigned int index, std::string* chr, int* beg, int* end) {
        assert(chr);
        chr->clear();
        int t = index; 
        int s; 
        for(unsigned int i = 0; i < this->chrVector.size(); i++ ) {
            s = rangeMap[chrVector[i]].size();
            if ( t < s) {
                (*chr) = (this->rangeMap[this->chrVector[i]][t]).chr;
                (*beg) = (this->rangeMap[this->chrVector[i]][t]).begin;
                (*end) = (this->rangeMap[this->chrVector[i]][t]).end;
                return;
            } else {
                t -= s;
            }
        }
        assert(false);
    };   
    size_t size() const { return this->rangeMap.size();};
    /**
     * check if position at chromosome @param c and position @param p
     * is in any stored range.
     * NOTE: using binary search could be faster (if PositionPair are sorted).
     */
    bool isInRange(const std::string& c, unsigned int p) {
        if (this->rangeMap.count(c) == 0) {
            return false;
        } 
        std::vector<PositionPair>& r = this->rangeMap[c];
        for (unsigned int i = 0; i < r.size(); i++) {
            if (r[i].isInRange(p)) {
                return true;
            }
        }
        return false;
    };
  private:
    void sortChrVector() {
        std::sort(chrVector.begin(), chrVector.end());
    };
    void dump(const std::vector<PositionPair>& v){
        for (unsigned int i = 0; i < v.size(); i++){
            printf("[%d, %d] ", v[i].begin, v[i].end);
        }
        printf("\n");
    }
    void consolidate(){
        std::map< std::string, std::vector<PositionPair> >::iterator iter;
        for (iter = this->rangeMap.begin(); iter != this->rangeMap.end(); iter++) {
            sortPositionRange( & ((*iter).second));
            // dump( ((*iter).second));
            consolidateRange(& ((*iter).second) );
            // dump( ((*iter).second));
        }
    };
    // sort range in ascending order
    void sortPositionRange(std::vector<PositionPair>* v) {
        std::sort(v->begin(), v->end(), PositionPairCompare);
    };
    // now the range in v should be ordered
    // we will merge overlapped ranges
    void consolidateRange(std::vector<PositionPair>* v) {
        std::vector<PositionPair> t;
        unsigned int l = v->size();
        if (l == 0) 
            return;

        unsigned int beg =  (*v)[0].begin;
        t.push_back( (*v)[0] );
        for(unsigned int i = 1; i < l; i++) {
            // if this range falls into last range, skip
            if ( (*v)[i].end <= (*v)[i - 1].end )
                continue;
            
            // if this range overlaps with last range, extend last range
            if ( (*v)[i].begin <= (*v)[i-1].end && 
                 (*v)[i].end > (*v)[i-1].end) {
                t[i-1].end = (*v)[i].end;
                continue;
            }

            // this range now is after the last range, so append it to the list
            t.push_back( (*v)[i] );
        }
        
        // copy t -> v
        v->clear();
        std::swap( t, *v);
    };
  private:
    std::vector<std::string> chrVector;
    std::map< std::string, std::vector<PositionPair> > rangeMap;
};

class RangeList{
public:
    RangeList(): isSorted(false) {};
    void obtainRange(const unsigned int index, std::string* range) {
        this->ensureSorted();
        this->rangeCollection.obtainRange(index, range);
    }
    void obtainRange(const unsigned int index, std::string* chr, int* beg, int* end) {
        this->ensureSorted();
        this->rangeCollection.obtainRange(index, chr, beg, end);
    }
    unsigned int size() const {return this->rangeCollection.size(); };
    // read gene list file and add these ranges
    void filterGeneName(const char* geneName, const char* fileName);
    void addRangeList(const char* argRangeList);
    void addRangeFile(const char* argRangeFile);
    void addRange(const std::string& id, const std::string& chr, unsigned int begin, unsigned int end) {
        this->isSorted = false;
        this->rangeCollection.addRange(id, chr, begin, end);
    };
private:
    void ensureSorted() {
        if (!this->isSorted) {
            this->rangeCollection.sort();
            this->isSorted = true;
        }
    };
    RangeCollection rangeCollection;
    bool isSorted;
};

//
void RangeList::filterGeneName(const char* inclusionGeneFileName, const char* geneTableFileName){
    String fileName = geneTableFileName;
    String geneName = inclusionGeneFileName;
    // require user input gene list file
    if (fileName.Length() == 0 && geneName.Length() != 0) {
        printf("Please provide gene list file (e.g. refFlat) until we are able to process\n");
        exit(1);
    }

    // if not specify range, return whole range.
    if (geneName.Length() == 0) {
        return;
    }

    // store which gene do we want if specified
    std::set< std::string > inclusionSet;
    inclusionSet.insert(geneName.c_str());
    
    String line;
    StringArray fields;
    std::string chr;
    std::string geneNameTbl;
    IFILE ifile = ifopen(fileName, "r");
    while (line.ReadLine(ifile) >= 0) {
        fields.Clear();
        fields.AddTokens(line, '\t');
        geneNameTbl = fields[0].c_str();
        if (inclusionSet.find(geneNameTbl) != inclusionSet.end()){ // store gene range
            chr = chopChr(fields[2].c_str());
            this->rangeCollection.addRange(std::string(line.c_str()), chr, 
                                     atoi(fields[4].c_str()),   // start
                                     atoi(fields[5].c_str()));   // end
        }
    }
    if (this->rangeCollection.size() == 0){
        fprintf(stdout, "We cannot find given gene in your geneListFile, so all sites will be outputed\n");
    }
}

// verify if s is of the format: chr:begin-end format
// @return true: valid format
bool verifyRangeFormat(const String& s, char* chr, unsigned int* begin, unsigned int* end) {
    int b = 0;
    int e = 0;
    while(s[e] != ':' && e < s.Length())
        e++;
    if (s[e] != ':') return false;
    strncpy(chr, s, e);
    chr[e]='\0';

    b = e + 1;
    e ++;
    std::string temp;
    while(s[e] != '-' && e < s.Length()){
        if (s[e] <'0' || s[e] > '9')
            return false;
        temp.push_back(s[e]);
        e++;
    }
    *begin = atoi(temp.c_str());

    if (s[e] != '-') return false;
    b = e + 1;
    e ++;
    temp = "";
    while(e < s.Length()){
        if (s[e] <'0' || s[e] > '9')
            return false;
        temp.push_back(s[e]);
        e++;
    }
    *end = atoi(temp.c_str());
    return true;
}

// input range such as: 1:100-200,3:200-300
void RangeList::addRangeList(const char* argRangeList) {
    String arg = argRangeList;
    StringArray col;
    col.AddTokens(arg, ',');
    for (int i = 0; i < col.Length(); i++){
        char c[64];
        unsigned int b,e;
        if (!verifyRangeFormat(col[i], c, &b, &e)) {
            fprintf(stdout, "This range does not conform 1:100-200 format -- %s\n", col[i].c_str());
        } else {
            this->rangeCollection.addRange("dummy", c, b, e);
        }
    }
};

/**
 * read a range list file like following
 * chr beg end
 * or 
 * chr beg
 * we will assume beg == end in the second case
 */
void RangeList::addRangeFile(const char* argRangeFile){
    IFILE ifile = ifopen(argRangeFile, "r");
    String line;

    while ( line.ReadLine(ifile) >= 0) {
        if (line.Length() == 0) continue;
    StringArray sa;

        sa.AddTokens(line);
        if (sa.Length() == 1){
            fprintf(stderr, "Wrong --rangeFile: %s\n", argRangeFile);
            return;
        } else if (sa.Length() == 2)
            this->rangeCollection.addRange(std::string(line.c_str()), std::string(sa[0].c_str()), (unsigned int) sa[1].AsInteger(), (unsigned int) sa[1].AsInteger());
        else if (sa.Length() == 3)
            this->rangeCollection.addRange(std::string(line.c_str()), std::string(sa[0].c_str()), (unsigned int) sa[1].AsInteger(), (unsigned int) sa[2].AsInteger());
        else {
            fprintf(stdout, "Will only use the first 3 column of --rangeFile %s\n", argRangeFile);
            this->rangeCollection.addRange(std::string(line.c_str()), sa[0].c_str(), (unsigned int) sa[1].AsInteger(), (unsigned int) sa[2].AsInteger());
        }
    }
};


#endif /* _RANGELIST_H_ */
